/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  ACCEM -- decompiles .BBS files into .MEC files                         *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#define ACCEM
#define INITIALIZE

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "prog.h"
#include "max_vr.h"
#include "mecca.h"

#define MAX_COMPILED  10

void pascal Process_Colour_Code(FILE *bbsfile,FILE *mecfile);
int sstsearch(char *key,struct _table base[],unsigned int num);
void pascal Get_Number_String(FILE *bbsfile,char *string,int max_len);
int cdecl trcmp(struct _table *v1,struct _table *v2);

long offsets[MAX_OFFSETS],
     fpos;
   
int num_offsets,
    llen;
    
int main(int argc,char *argv[])
{
  struct _inf inf;
  
  FILE *bbsfile,
       *mecfile;
     
  char string[MAX_COMPILED+5],
       temp[PATHLEN],
       inname[PATHLEN],
       outname[PATHLEN],
       *s;
  
  long save_fp,
       tl;

  int verbnum,
      split,
      iter,
      ch,
      x;



  Hello("ACCEM","The Maximus Embedded Command Decompiler",VERSION,"1990");
  
  if (argc < 2)
  {
    printf("Usage: ACCEM <infile> [outfile] [-s]\n\n");
    
    printf("If no extension is specified for <infile>, then ACCEM will default to\n");
    printf("an extension of .BBS.  Likewise, if no extension is specified for\n");
    printf("[outfile], then ACCEM will default to an extension of .MEC.  If no\n");
    printf("[outfile] is specified, then ACCEM will use the root of the input\n");
    printf("filename, with an extension of .MEC.\n\n");

    printf("The `-s' command-line parameter is optional: if used, it will cause\n");
    printf("ACCEM to split any lines which are over 100 characters in length, by\n");
    printf("placing an empty brace at the end of the line.  MECCA will ignore this\n");
    printf("brace when compiling, and act as if the two lines were one.  However,\n");
    printf("this switch may be necessary if your editor cannot handle lines over a\n");
    printf("certain length.\n");

    return 1;
  }

  split=FALSE;

  strcpy(inname,argv[1]);

  if (argc==3 && (eqstri(argv[2],"-s") || eqstri(argv[2],"/s")))
  {
    split=TRUE;
    argc--;
  }

  if (argc >= 3)
  {
    strcpy(outname,argv[2]);

    if (argc >= 4 && (eqstri(argv[3],"-s") || eqstri(argv[3],"/s")))
      split=TRUE;
  }
  else
  {
    strcpy(outname,inname);

    if ((s=strchr(outname,'.')) != NULL)
      *s='\0';
  }
  
  if (! strchr(inname,'.'))
    strcat(inname,".BBS");
  
   if (! strchr(outname,'.'))
    strcat(outname,".MEC");

  strupr(inname);
  strupr(outname);
    
  iqsort((char *)verbs,verb_table_size,sizeof(verbs[0]),trcmp);
  Init_Table();

  if ((bbsfile=fopen(inname,"rb"))==NULL)
  {
    printf("Error opening `%s' for read!\n",argv[1]);
    exit(1);
  }
  
  if ((mecfile=fopen(outname,"wb"))==NULL)
  {
    printf("Error opening `%s' for write!\n",argv[2]);
    exit(1);
  }

  num_offsets=llen=0;

  printf("Compiling `%s' to `%s':\n\n",inname,outname);

  printf("Pass 1: Building jump table\n");
  
  for (ch=0;ch != EOF;)
  {
    if ((ch=fgetc(bbsfile))=='\x0f' && (ch=fgetc(bbsfile))=='V')
    {
      Get_Number_String(bbsfile,string,5);
      tl=atol(string);

      for (x=0;x < num_offsets;x++)
        if (offsets[x]==tl)
          break;

      if (x==num_offsets) /* If's a new label offset (not found in table) */
        offsets[num_offsets++]=tl;
    }
  }

  printf("Pass 2: Decompiling\n");
  
  fseek(bbsfile,0L,SEEK_SET);
  fpos=0L;
  
  while (1)
  {
    save_fp=fpos;

    for (x=0;x < num_offsets;x++)
      if (fpos==offsets[x])
      {
        sprintf(temp,"[/L%d]",x);
        llen += strlen(temp);
        fputs(temp,mecfile);
        break;
      }

    if ((ch=fgetc(bbsfile))==EOF)
      break;

    if (llen >= MAX_LLEN)
    {
      if (split)
        fprintf(mecfile,"[\n]");

      llen=1;
    }

    fpos++;
    
    if (ch=='[')
    {
      fputs("[[",mecfile);
      llen += 2;
    }
    else if (ch >= 32 || ch=='\t' || ch=='\r' || ch=='\n')
    {
      fputc(ch,mecfile);
      llen++;

      if (ch=='\r' || ch=='\n')
        llen=0;
    }
    else
    {
      *string='\0';

      switch (ch)
      {
        case 22:
          fpos++;
          
          if ((ch=fgetc(bbsfile))=='\x01') /* Colour code */
          {
            Process_Colour_Code(bbsfile,mecfile);
            break;
          }
          else strcpy(string,"\x16");
          /* else fall-through, a normal AVATAR command */
        
        default:
          for (iter=0;iter < MAX_COMPILED;iter++)
          {
            string[x=strlen(string)]=(char)ch;
            string[x+1]='\0';

            if ((verbnum=sstsearch(string,verbs,verb_table_size)) != -1)
            {
              if (verbs[verbnum].fptr)
              {
                inf.p=NULL;
                inf.outfile=&mecfile;
                inf.infile=bbsfile;
                inf.outname=NULL;
                inf.verbnum=verbnum;
                
                (*verbs[verbnum].fptr)(&inf);
              }
              else
              {
                fprintf(mecfile,"[%s]",verbs[verbnum].verb);
                llen += 2+strlen(verbs[verbnum].verb);
              }
              break;
            }

            fpos++;
            
            if ((ch=fgetc(bbsfile))==EOF)
              break;
          }

          if (iter != MAX_COMPILED)
            break;

          sprintf(temp,"[%d]",(unsigned char)*string);
          fputs(temp,mecfile);
          llen += strlen(temp);
          
          /* Now skip over the bad char, and try again */
          fseek(bbsfile,fpos=save_fp+1,SEEK_SET);
          continue;
      }
    }
  }
  
  printf("\nDone!\n");
  
  fclose(mecfile);
  fclose(bbsfile);

  return 0;
}

int sstsearch(char *key,struct _table base[],unsigned int num)
{
  int x,
      lastx=-1,
      lasthi,
      lastlo;

  char *s,*t;

  lasthi=num;
  lastlo=0;

  while (1)
  {
    x=((lasthi-lastlo) >> 1)+lastlo;

    if (lastx==x)
      return -1;

    lastx=x;

    for (s=key,t=base[x].translation;*s==*t;s++,t++)
      if (! *s)
        return (x);                       /* Found a match */

    if (*s > *t)
      lastlo=x;
    else
      lasthi=x;
  }

  return -1;
}


void pascal Process_Colour_Code(FILE *bbsfile,FILE *mecfile)
{
  int ch;
  char temp[PATHLEN];
  
  /* Unescape any control codes with their high bit set. */
  
  fpos++;
  
  if ((ch=fgetc(bbsfile))==DLE)
  {
    ch=fgetc(bbsfile) & 0x7f;
    fpos++;
  }

  colour_to_string(ch,temp);
  fprintf(mecfile,"[%s]",temp);
  llen += 2+strlen(temp);
}



int cdecl trcmp(struct _table *v1,struct _table *v2)
{
  return (strcmp(v1->translation,v2->translation));
}

      
void pascal Get_Number_String(FILE *bbsfile,char *string,int max_len)
{
  int ch,
      x;
  
  *string='\0';
  
  while ((ch=fgetc(bbsfile)) >= '0' && ch <= '9')
  {
    if ((x=strlen(string))==max_len)
      break;

    string[x]=(char)ch;
    string[x+1]='\0';

    fpos++;
  }

  ungetc(ch,bbsfile);
}


void pascal P_Comment (struct _inf *inf) {(void)inf;}
void pascal P_Copy    (struct _inf *inf) {(void)inf;}
void pascal P_Include (struct _inf *inf) {(void)inf;}
void pascal P_On      (struct _inf *inf) {(void)inf;}
void pascal P_Save    (struct _inf *inf) {(void)inf;}
void pascal P_Load    (struct _inf *inf) {(void)inf;}
void pascal P_Fg      (struct _inf *inf) {(void)inf;}
void pascal P_Bg      (struct _inf *inf) {(void)inf;}
void pascal P_Steady  (struct _inf *inf) {(void)inf;}
void pascal P_Dim     (struct _inf *inf) {(void)inf;}
void pascal P_Bright  (struct _inf *inf) {(void)inf;}
void pascal P_Priv    (struct _inf *inf) {(void)inf;}
void pascal P_Label   (struct _inf *inf) {(void)inf;}
void pascal P_Restore (struct _inf *inf) {(void)inf;}

void pascal P_Locate(struct _inf *inf)
{
  int ch;
  char temp[PATHLEN];
  
  fpos += 2;
  
  ch=fgetc(inf->infile);
  
  sprintf(temp,"[locate %2d %2d]",ch,fgetc(inf->infile));
  fputs(temp,*inf->outfile);
  llen += strlen(temp);
}

void pascal P_Menu_Cmd(struct _inf *inf)
{
  int ch;
  char temp[PATHLEN];
  
  fpos++;
  
  ch=fgetc(inf->infile);
  
  sprintf(temp,"[menu_cmd %2d]",ch);
  fputs(temp,*inf->outfile);
  llen += strlen(temp);
}


void pascal P_Setpriv (struct _inf *inf)
{
  char temp[5], 
       *s;
  
  int ch,
      x;
    
  
  fpos++;
  ch=fgetc(inf->infile);
  
  temp[0]=(char)toupper(ch);
  temp[1]='\0';

  for (x=0,s=temp;x < _PRIVS_NUM;x++)
    if (ch==*_privs[x].name)
    {
      s=_privs[x].name;
      break;
    }
    
  fprintf(*inf->outfile,"[setpriv %s]",s);
  llen += 2+strlen(s);
}

void pascal P_Iftime  (struct _inf *inf)
{
  static char *time_verbs[]={"GT","LT","EQ","NE","GE","LE",NULL};

  char temp[PATHLEN],
       *s;
  
  int ch,
      x;
  
  fpos++;
  ch=fgetc(inf->infile)-1;
  
  for (x=0,s=NULL;time_verbs[x];x++)
    if (toupper(ch)==x)
    {
      s=time_verbs[x];
      break;
    }
    
  if (! s)
    s=time_verbs[0];
  
  fpos++;
  ch=fgetc(inf->infile);
  
  fpos++;
  
  sprintf(temp,"[iftime %s %02d:%02d]",s,ch-1,fgetc(inf->infile)-1);
  fputs(temp,*inf->outfile);
  llen += strlen(temp);
}


void pascal P_Repeat(struct _inf *inf)
{
  int ch,
      x,n;

  fpos += 2;
  ch=fgetc(inf->infile);
  n=fgetc(inf->infile);

  for (x=0;x < n;x++)
  {
    if (ch=='[')  /* Double up any left brackets */
    {
      fputc('[',*inf->outfile);
      llen++;
    }

    fputc(ch,*inf->outfile);
    llen++;
  }
}

void pascal P_Repeatseq(struct _inf *inf)
{
  char temp[PATHLEN];

  int ch;

  sprintf(temp,"[repeatseq %d",ch=fgetc(inf->infile));

  /* Increase fpos by number of chars, plus length, plus repeat times */
  fpos += ch+2;

  /* Increase ch by one, to grab the #-of-repeats byte */
  ch++;

  while (ch--)
    sprintf(&temp[strlen(temp)]," %d",fgetc(inf->infile));

  strcat(temp,"]");

  llen += strlen(temp);

  fputs(temp,*inf->outfile);
}




void pascal P_Goto(struct _inf *inf)
{
  char string[6],
       temp[PATHLEN];
  long tl;
  int x;

  Get_Number_String(inf->infile,string,5);
  tl=atol(string);

  for (x=0;x < num_offsets;x++)
    if (tl==offsets[x])
    {
      sprintf(temp,"[goto L%d]",x);
      fputs(temp,*inf->outfile);
      llen += strlen(temp);
      break;
    }

  if (x==num_offsets)
  {
    fprintf(*inf->outfile,"[goto Unknown]");
    llen += 14;
  }
}


